// TestWnd.cpp : implementation file
//

#include "stdafx.h"
#include "DDEval.h"
#include "TestWnd.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CTestWnd

CTestWnd::CTestWnd()
{
	m_pTI = NULL;
	m_pDD = NULL;
}

CTestWnd::~CTestWnd()
{
	delete m_pDD;
	delete m_pTI;
}


BEGIN_MESSAGE_MAP(CTestWnd, CWnd)
	//{{AFX_MSG_MAP(CTestWnd)
	ON_WM_CHAR()
	ON_WM_DESTROY()
	ON_WM_ERASEBKGND()
	ON_WM_TIMER()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////

// Create the test window
BOOL CTestWnd::Create(TESTINFO* pti)
{
	// Save test info
	m_pTI = pti;
	ASSERT(m_pTI);

	// Create a DirectDraw object
	m_pDD = new CDirectDraw;
	BOOL b = m_pDD->Create();
	ASSERT(b);
	
	// Register a class
	CString strClass =
		AfxRegisterWndClass(CS_HREDRAW | CS_VREDRAW,
							::LoadCursor(NULL, IDC_ARROW),
							(HBRUSH)::GetStockObject(GRAY_BRUSH));

	// Define window style and size
	DWORD dwStyle = WS_VISIBLE | WS_POPUP;
	RECT rc;
	if (m_pTI->bFullScreen) {
		rc.top = 0;
		rc.left = 0;
		rc.right = ::GetSystemMetrics(SM_CXSCREEN);
		rc.bottom = ::GetSystemMetrics(SM_CYSCREEN);
	} else { // Windowed
		dwStyle |= WS_CAPTION | WS_SYSMENU;
		rc.top = 50;
		rc.left = 50;
		rc.bottom = rc.top + m_pTI->iHeight;
		rc.right = rc.left + m_pTI->iWidth;
		::AdjustWindowRect(&rc, dwStyle, FALSE);
	}

	if (!CreateEx(0,
				  strClass,
				  "DirectDraw Window",
				  dwStyle,
				  rc.left, rc.top,
				  rc.right - rc.left, rc.bottom - rc.top,
				  m_pTI->pParent->GetSafeHwnd(),
				  NULL)) {
		return FALSE;
	}

	// Ensure window is visible
	UpdateWindow();

	// Set mode for window
	ASSERT(m_pTI);
	if (m_pTI->bFullScreen) {
		b = m_pDD->SetFullScreenMode(GetSafeHwnd(),
									 m_pTI->iWidth,
									 m_pTI->iHeight,
									 m_pTI->iBpp);
	} else {
		b = m_pDD->SetWindowedMode(GetSafeHwnd(),
								   m_pTI->iWidth,
								   m_pTI->iHeight);
	}
	ASSERT(b);

	// Run tests
	SetTimer(1, 100, NULL);
	return TRUE;
}

void CTestWnd::OnTimer(UINT nIDEvent) 
{
	// Run next test
	switch (m_pTI->iTest) {
	case 1:
		TestGDIText();
		break;
	case 2:
		TestGDIGfx();
		break;
	case 3:
		TestDDSprite();
		break;
	case 4:
		TestDirectPixels();
		break;
	default:
		ASSERT(0);
		break;
	}
}

void CTestWnd::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// Any key ends the test
	DestroyWindow();
}

void CTestWnd::OnDestroy() 
{
	// Stop the update timer
	KillTimer(1);

	// Restore the screen mode
	ASSERT(m_pTI);
	if (m_pTI->bFullScreen) {
		m_pDD->RestoreMode();
	}
	CWnd::OnDestroy();
}

void CTestWnd::PostNcDestroy() 
{
	delete this;
}

// Disable the default background erase
BOOL CTestWnd::OnEraseBkgnd(CDC* pDC) 
{
	return TRUE;
}

void CTestWnd::TestGDIText()
{
	ASSERT(m_pTI);

	// Get front and back buffers
	CDDSurface* pBB = m_pDD->GetBackBuffer();
	ASSERT(pBB);
	CDDSurface* pFB = m_pDD->GetFrontBuffer();
	ASSERT(pFB);

	// Get rectangle that describes front buffer
	RECT rcFront;
	if (m_pTI->bFullScreen) {
		pFB->GetRect(rcFront);
	} else {
		GetClientRect(&rcFront);
		ClientToScreen(&rcFront);
	}

	RECT rcBack;
	pBB->GetRect(rcBack);

	DWORD dwStart = timeGetTime();
	int nFrames = 100;
	for (int iFrame = 0; iFrame < nFrames; iFrame++) {

		DWORD dwNow = timeGetTime();
		double fps; 
		if (dwNow == dwStart) {
			fps = 0;
		} else {
			fps = iFrame * 1000.0 / (double)(dwNow - dwStart);
		}

		// Make up some text
		char buf[64];
		sprintf(buf, "Frame %d  (%3.1f fps)", iFrame, fps);

		// Get a DC to back buffer
		CDC* pdc = pBB->GetDC();
		ASSERT(pdc);

		// Fill buffer with white
		pdc->PatBlt(rcBack.left,
					rcBack.top,
					rcBack.right - rcBack.left,
					rcBack.bottom - rcBack.top,
					WHITENESS);

		// Draw text
		pdc->DrawText(buf,
					  -1,
					  &rcBack,
					  DT_CENTER | DT_BOTTOM | DT_SINGLELINE);

		// Release DC
		pBB->ReleaseDC(pdc);

		// Swap buffers
		if (m_pTI->bFullScreen) {
			pFB->Flip();
		} else {
			pFB->Blt(&rcFront, pBB, &rcBack);
		}
	}
}

void CTestWnd::TestGDIGfx()
{
	ASSERT(m_pTI);

	// Get the front and back buffers
	CDDSurface* pBB = m_pDD->GetBackBuffer();
	ASSERT(pBB);
	CDDSurface* pFB = m_pDD->GetFrontBuffer();
	ASSERT(pFB);

	// Get the rectangle that describes the front buffer
	RECT rcFront;
	if (m_pTI->bFullScreen) {
		pFB->GetRect(rcFront);
	} else {
		GetClientRect(&rcFront);
		ClientToScreen(&rcFront);
	}

	RECT rcBack;
	pBB->GetRect(rcBack);

	static int x = 0;
	static int y = 0;
	static int vx = 5;
	static int vy = 5;
	int cx = 100;
	int cy = 100;
	
	DWORD dwStart = timeGetTime();
	int nFrames = 100;
	for (int iFrame = 0; iFrame < nFrames; iFrame++) {

		DWORD dwNow = timeGetTime();
		double fps; 
		if (dwNow == dwStart) {
			fps = 0;
		} else {
			fps = iFrame * 1000.0 / (double)(dwNow - dwStart);
		}
		char buf[64];
		sprintf(buf, "Frame %d  (%3.1f fps)", iFrame, fps);

		// Get a DC to the back buffer
		CDC* pdc = pBB->GetDC();
		ASSERT(pdc);

		// Fill the buffer with white
		pdc->PatBlt(rcBack.left,
					rcBack.top,
					rcBack.right - rcBack.left,
					rcBack.bottom - rcBack.top,
					WHITENESS);

		// Draw the picture
		pdc->Rectangle(x, y, x+cx, y+cy);

		// Draw the text
		pdc->DrawText(buf,
					  -1,
					  &rcBack,
					  DT_CENTER | DT_BOTTOM | DT_SINGLELINE);

		// Finished with the DC now
		pBB->ReleaseDC(pdc);

		// Swap buffers
		if (m_pTI->bFullScreen) {
			pFB->Flip();
		} else {
			pFB->Blt(&rcFront, pBB, &rcBack);
		}

		// Advance the position
		x += vx;
		if (vx > 0) {
			if (x+cx > rcBack.right) {
				vx = -vx;
				x = rcBack.right - cx;
			}
		} else {
			if (x < 0) {
				vx = -vx;
				x = 0;
			}
		}
		y += vy;
		if (vy > 0) {
			if (y+cy > rcBack.bottom) {
				vy = -vy;
				y = rcBack.bottom - cy;
			}
		} else {
			if (y < 0) {
				vy = -vy;
				y = 0;
			}
		}
	}
}

void CTestWnd::TestDDSprite()
{
	ASSERT(m_pTI);

	// Get the front and back buffers
	CDDSurface* pBB = m_pDD->GetBackBuffer();
	ASSERT(pBB);
	CDDSurface* pFB = m_pDD->GetFrontBuffer();
	ASSERT(pFB);

	// Get the rectangle that describes the front buffer
	RECT rcFront;
	if (m_pTI->bFullScreen) {
		pFB->GetRect(rcFront);
	} else {
		GetClientRect(&rcFront);
		ClientToScreen(&rcFront);
	}
	RECT rcBack;
	pBB->GetRect(rcBack);

	// Create the sprite in a surface of its own
	static int x = 0;
	static int y = 0;
	static int vx = 5;
	static int vy = 5;
	int cx = 100;
	int cy = 100;

	DDSURFACEDESC sd;
	memset(&sd, 0, sizeof(sd));
	sd.dwSize = sizeof(sd);
	sd.dwFlags = DDSD_WIDTH 
			   | DDSD_HEIGHT
			   | DDSD_CAPS;
	sd.dwWidth = cx;
	sd.dwHeight = cy;
	sd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	CDDSurface sprite;
	BOOL b = sprite.Create(m_pDD, &sd);
	ASSERT(b);
	
	// Get DC to sprite surface and draw sprite
	// as a red circle on a black background
	CDC* pdc = sprite.GetDC();
	ASSERT(pdc);
	pdc->PatBlt(0, 0, cx, cy, BLACKNESS);
	CBrush br;
	br.CreateSolidBrush(RGB(255, 0, 0));
	CBrush* pbrOld = pdc->SelectObject(&br);
	pdc->Ellipse(0, 0, cx, cy);
	pdc->SelectObject(pbrOld);
	sprite.ReleaseDC(pdc);

	// Set color key as black
	DDCOLORKEY ck;
	ck.dwColorSpaceLowValue = 0; // Black
	ck.dwColorSpaceHighValue = 0;
	sprite.SetColorKey(DDCKEY_SRCBLT, &ck);
	
	DWORD dwStart = timeGetTime();
	int nFrames = 100;
	for (int iFrame = 0; iFrame < nFrames; iFrame++) {

		DWORD dwNow = timeGetTime();
		double fps; 
		if (dwNow == dwStart) {
			fps = 0;
		} else {
			fps = iFrame * 1000.0 / (double)(dwNow - dwStart);
		}
		char buf[64];
		sprintf(buf, "Frame %d  (%3.1f fps)", iFrame, fps);

		// Fill buffer with white
		CDC* pdc = pBB->GetDC();
		ASSERT(pdc);
		pdc->PatBlt(rcBack.left,
					rcBack.top,
					rcBack.right - rcBack.left,
					rcBack.bottom - rcBack.top,
					WHITENESS);

		// Draw text
		pdc->DrawText(buf,
					  -1,
					  &rcBack,
					  DT_CENTER | DT_BOTTOM | DT_SINGLELINE);

		pBB->ReleaseDC(pdc);

		// Draw sprite
		RECT rcDst;
		rcDst.left = x;
		rcDst.top = y;
		rcDst.right = rcDst.left + cx;
		rcDst.bottom = rcDst.top + cy;
		RECT rcSrc;
		rcSrc.left = 0;
		rcSrc.top = 0;
		rcSrc.right = rcSrc.left + cx;
		rcSrc.bottom = rcSrc.top + cy;

		// Call Blt using a color key
		pBB->Blt(&rcDst, &sprite, &rcSrc,
				 DDBLT_WAIT|DDBLT_KEYSRC);

		// Swap buffers
		if (m_pTI->bFullScreen) {
			pFB->Flip();
		} else {
			pFB->Blt(&rcFront, pBB, &rcBack);
		}

		// Advance the position
		x += vx;
		if (vx > 0) {
			if (x+cx > rcBack.right) {
				vx = -vx;
				x = rcBack.right - cx;
			}
		} else {
			if (x < 0) {
				vx = -vx;
				x = 0;
			}
		}
		y += vy;
		if (vy > 0) {
			if (y+cy > rcBack.bottom) {
				vy = -vy;
				y = rcBack.bottom - cy;
			}
		} else {
			if (y < 0) {
				vy = -vy;
				y = 0;
			}
		}
	}
}

void NextColor(COLORREF& clr)
{
	int r = GetRValue(clr);
	int g = GetGValue(clr);
	int b = GetBValue(clr);
	int d = 4;
	if (r > 0) {
		r += d;
		if (r > 255) {
			r = 0;
			g = d;
		}
	} else if (g > 0) {
		g += d;
		if (g > 255) {
			g = 0;
			b = d;
		}
	} else if (b > 0) {
		b += d;
		if (b > 255) {
			b = 0;
			r = d;
		}
	} else {
		r = d;
	}
	clr = RGB(r, g, b);
}

void CTestWnd::TestDirectPixels()
{
	ASSERT(m_pTI);

	// Get the front and back buffers
	CDDSurface* pBB = m_pDD->GetBackBuffer();
	ASSERT(pBB);
	CDDSurface* pFB = m_pDD->GetFrontBuffer();
	ASSERT(pFB);

	// Get the rectangle that describes the front buffer
	RECT rcFront;
	if (m_pTI->bFullScreen) {
		pFB->GetRect(rcFront);
	} else {
		GetClientRect(&rcFront);
		ClientToScreen(&rcFront);
	}
	RECT rcBack;
	pBB->GetRect(rcBack);

	// Get some info about back buffer
	int iBpp = pBB->GetBitsPerPixel();
	ASSERT(iBpp >= 8);
	int iWidth = pBB->GetWidth();
	int iHeight = pBB->GetHeight();
	int iPitch = pBB->GetPitch();

	// Get RGB masks
	DWORD dwRMask, dwGMask, dwBMask;
	pBB->GetRGBMasks(dwRMask, dwGMask, dwBMask);

	// For each mask, check its width in bits
	// and see how many bits from the LSB (lease significant
	// bit) end it's shifted
	DWORD dwRShift, dwGShift, dwBShift;
	DWORD dwRBits, dwGBits, dwBBits;
	dwRShift = dwRBits = 0;
	dwGShift = dwGBits = 0;
	dwBShift = dwBBits = 0;
	if (iBpp > 8) {
		DWORD d = dwRMask;
		while ((d & 0x1) == 0) {
			d = d >> 1;
			dwRShift++;
		}
		while (d & 0x01) {
			d = d >> 1;
			dwRBits++;
		}
		d = dwGMask;
		while ((d & 0x1) == 0) {
			d = d >> 1;
			dwGShift++;
		}
		while (d & 0x01) {
			d = d >> 1;
			dwGBits++;
		}
		d = dwBMask;
		while ((d & 0x1) == 0) {
			d = d >> 1;
			dwBShift++;
		}
		while (d & 0x01) {
			d = d >> 1;
			dwBBits++;
		}
	}
	
	// Create a color to use for the lines. We'll update this
	// every frame so the image will constantly change
	static COLORREF clrStart = RGB(0, 0, 0);

	// If buffer is 8bpp, get palette entries
	// and set them to color values we want to use
	PALETTEENTRY pe[256];
	BYTE r, g, b;
	CDDPalette* pPal = pBB->GetPalette();
	if (pPal) {
		pPal->GetEntries(0, 256, pe);

		// Set color values we want to use.
		// We'll create a small color cube using 2 bits for
		// R, G and B
		for (r = 0; r < 4; r++) {
			for (g = 0; g < 4; g++) {
				for (b = 0; b < 4; b++) {
					int index = 10 + r * 16 + g * 4 + b;
					pe[index].peRed = r * 85;
					pe[index].peGreen = g * 85;
					pe[index].peBlue = b * 85;
				}
			}
		}
		
		// Fill remainder of palette entries with gray just
		// so we can see them while debugging
		for (int i = 10 + 4*4*4; i < 246; i++) {
			pe[i].peRed = 192;
			pe[i].peGreen = 192;
			pe[i].peBlue = 192;
		}

		// Update palette
		pPal->SetEntries(0, 256, pe);

		// Delete palette
		delete pPal;
	}

	DWORD dwStart = timeGetTime();
	int nFrames = 100;
	for (int iFrame = 0; iFrame < nFrames; iFrame++) {

		DWORD dwNow = timeGetTime();
		double fps; 
		if (dwNow == dwStart) {
			fps = 0;
		} else {
			fps = iFrame * 1000.0 / (double)(dwNow - dwStart);
		}
		char buf[64];
		sprintf(buf, "Frame %d  (%3.1f fps)", iFrame, fps);

		// Fill the buffer with white
		// GDI is fast for this!
		CDC* pdc = pBB->GetDC();
		ASSERT(pdc);
		pdc->PatBlt(rcBack.left,
					rcBack.top,
					rcBack.right - rcBack.left,
					rcBack.bottom - rcBack.top,
					WHITENESS);
		pBB->ReleaseDC(pdc);

		// Draw the picture
		COLORREF clrLine = clrStart;

		// Lock buffer and get pointer.
		// WARNING: Don't try stepping until Unlock.
		BYTE* pBuf = (BYTE*) pBB->Lock();
		if (pBuf) {

			for (int y = 0; y < iHeight; y++) {

				// Compute where line segment starts
				int n = iWidth;
				DWORD dwOffset = y * iPitch; // In bytes
				
				// Get color
				int ir = GetRValue(clrLine);
				int ig = GetGValue(clrLine);
				int ib = GetBValue(clrLine);

				// Draw pixels directly to buffer
				switch (iBpp) {
				case 8: {
					// Find index into color cube for this
					// color
					int index = 10 + (ir / 85) * 16 + (ig / 85) * 4 + (ib / 85);

					BYTE* p = pBuf + dwOffset;
					while (n--) {
						*p++ = (BYTE) index;
					}
					} break;

				case 16: {
					// Build color
					DWORD dw = (ir >> (8 - dwRBits)) << dwRShift
							 | (ig >> (8 - dwGBits)) << dwGShift
							 | (ib >> (8 - dwBBits)) << dwBShift;
					WORD w = (WORD)dw;
					WORD* p = (WORD*)(pBuf + dwOffset);
					while (n--) *p++ = w;
					} break;

				case 24:
					// Student exercise :)
					break;

				case 32: {
					DWORD dw = (ir >> (8 - dwRBits)) << dwRShift
							 | (ig >> (8 - dwGBits)) << dwGShift
							 | (ib >> (8 - dwBBits)) << dwBShift;
					DWORD* p = (DWORD*)(pBuf + dwOffset);
					while (n--) *p++ = dw; 
					} break;

				default:
					break;
				}

				// Move on to next color
				NextColor(clrLine);

			}
			pBB->Unlock();
			// You can debug again now
		}
		NextColor(clrStart);
		
		// Draw the text
		pdc = pBB->GetDC();
		ASSERT(pdc);
		pdc->DrawText(buf,
					  -1,
					  &rcBack,
					  DT_CENTER | DT_BOTTOM | DT_SINGLELINE);
		pBB->ReleaseDC(pdc);

		// Swap buffers
		if (m_pTI->bFullScreen) {
			pFB->Flip();
		} else {
			pFB->Blt(&rcFront, pBB, &rcBack);
		}
	}
}

